@@ -7,9 +7,9 @@
 __package_info__ = """
 BEGIN PYTHON-PACKAGE-INFO 1.0
 Title:			mxTextTools - Tools for fast text processing
-Current-Version:	1.0.2
+Current-Version:	1.1.1
 Home-Page:		http://starship.skyport.net/~lemburg/mxTextTools.html
-Primary-Site:		http://starship.skyport.net/~lemburg/mxTextTools-1.0.2.zip
+Primary-Site:		http://starship.skyport.net/~lemburg/mxTextTools-1.1.1.zip
 
 This package provides several different functions and mechanisms
 to do fast text text processing. Amongst these are character set
@@ -20,3 +20,29 @@ END PYTHON-PACKAGE-INFO
 """
 from TextTools import *
 from TextTools import __version__
+
+### Make the types pickleable:
+
+# Shortcuts for pickle (reduces the pickle's length)
+def _BMS(match,translate):
+    return BMS(match,translate)
+def _FS(match,translate):
+    return FS(match,translate)
+
+# Module init
+class modinit:
+
+    ### Register the two types
+    import copy_reg
+    def pickle_BMS(so):
+	return _BMS,(so.match,so.translate)
+    def pickle_FS(so):
+	return _FS,(so.match,so.translate)
+    copy_reg.pickle(BMSType,
+		    pickle_BMS,
+		    _BMS)
+    copy_reg.pickle(FSType,
+		    pickle_FS,
+		    _FS)
+
+del modinit
