@@ -6,3 +6,12 @@
 """
 from mxTextTools import *
 from mxTextTools import __version__
+
+#
+# Make BMS take the role of FS in case the Fast Search object was not built
+#
+try:
+    FS
+except NameError:
+    FS = BMS
+    FSType = BMSType
