@@ -16,6 +16,10 @@
 
 */
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /* sanity check switches */
 /*#define SAFER 1*/
 
@@ -51,4 +55,7 @@ extern int bm_tr_search(mxbmse_data *c,
 			char *tr);
 
 /* EOF */
+#ifdef __cplusplus
+}
+#endif
 #endif
