@@ -120,18 +120,18 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 	IF_DEBUGGING {
 	    PyObject *v;
 	    
-	    DPRINTF("\n# Tag Table entry %i: ",i);
+	    mxDebugPrintf("\n# Tag Table entry %i: ",i);
 	    v = PyObject_Repr(tagobj);
 	    if (!v) {
 		v = tagobj;
 		Py_INCREF(v);
 	    }
 	    if (PyString_Check(v))
-		DPRINTF("tagobj=%s cmd=%i flags=%i position=%i\n",
-			PyString_AsString(v),cmd,flags,x);
+		mxDebugPrintf("tagobj=%s cmd=%i flags=%i position=%i\n",
+			      PyString_AsString(v),cmd,flags,x);
 	    else
-		DPRINTF("tagobj at 0x%lx cmd=%i flags=%i position=%i\n",
-			(long)v,cmd,flags,x);
+		mxDebugPrintf("tagobj at 0x%lx cmd=%i flags=%i position=%i\n",
+			      (long)v,cmd,flags,x);
 	    Py_DECREF(v);
 	}
 	
@@ -409,7 +409,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 	
 	    switch (cmd) {
 
-	    case MATCH_FAIL:
+	    case MATCH_FAIL: /* == MATCH_JUMP */
 
 		if (jne == 0) { /* match failed */
 		    rc = 1;
@@ -617,6 +617,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 	    }
 	    
 	case MATCH_TABLE:
+	case MATCH_SUBTABLE:
 	    
 	    if (PyInt_Check(match) && PyInt_AS_LONG(match) == MATCH_THISTABLE)
 		match = table;
@@ -626,26 +627,26 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 		int y = x;
 		int newrc = 0;
 
-		if (taglist != Py_None) {
-		    /* want a list, so create one */
+		if (taglist != Py_None && cmd != MATCH_SUBTABLE) {
+		    /* Create a new list for use as subtaglist */
 		    subtags = PyList_New(0);
 		    if (subtags == NULL) 
 			goto onError;
 		}
-		else { 
-		    /* better without list */
-		    subtags = Py_None;
+		else {
+		    /* Use taglist as subtaglist */
+		    subtags = taglist;
 		    Py_INCREF(subtags);
 		}
 
-		DPRINTF("\nTable : using table at 0x%lx\n",(long)match);
+		DPRINTF("\n[Sub]Table : using table at 0x%lx\n",(long)match);
 
 		start = x;
 
 		/* match other table */
 		newrc = fast_tag(pytext,text,len_text,match,start,subtags,&y);
 		if (newrc == 0) {
-		    Py_MY_DELETE(subtags);
+		    Py_DECREF(subtags);
 		    goto onError;
 		}
 
@@ -655,6 +656,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 		    if (jne == 0) {
 			/* match failed */
 			rc = 1; 
+			Py_DECREF(subtags);
 			goto finished;
 		    }
 		    else 
@@ -670,84 +672,27 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 			if (match_append(flags,pytext,taglist,tagobj,
 					 start,x,subtags) < 0) {
 			    /* append failed */
-			    Py_MY_DELETE(subtags);
+			    Py_DECREF(subtags);
 			    goto onError;
-			DPRINTF(" [%i:%i] (matched and remembered this slice)\n",
-				start,x);
+			    DPRINTF(" [%i:%i] (matched and remembered this slice)\n",
+				    start,x);
 			}
 		    }
 		    else
 			DPRINTF(" [%i:%i] (matched but not saved)\n",start,x);
 		}
-
-		Py_MY_DELETE(subtags);
+		Py_DECREF(subtags);
 		goto next_entry;
 	    }
 	    else
 		Py_ErrorWithArg(PyExc_TypeError,
 				"Tag Table entry %i: "
-				"expected a table (cmd=Table)",i);
-
-	    goto next_entry;
-	    
-	case MATCH_SUBTABLE:
-	    
-	    if (PyInt_Check(match) && PyInt_AS_LONG(match) == MATCH_THISTABLE)
-		match = table;
-
-	    if (PyTuple_Check(match)) { 
-		int y = x;
-		int newrc = 0;
-
-		DPRINTF("\nSubTable : using table at 0x%lx\n",(long)match);
-
-		start = x;
-
-		/* match other table */
-		newrc = fast_tag(pytext,text,len_text,match,start,taglist,&y);
-		if (newrc == 0) {
-		    goto onError;
-		}
-
-		if (newrc == 1) {
-		    /* not matched */
-		    DPRINTF(" (no success)\n");
-		    if (jne == 0) {
-			/* match failed */
-			rc = 1; 
-			goto finished;
-		    }
-		    else 
-			je = jne;
-		} 
-		else { 
-		    /* matched */
-
-		    /* move x to new position */
-		    x = y;
-
-		    if (tagobj != Py_None) {
-			if (match_append(flags,pytext,taglist,tagobj,
-					 start,x,NULL) < 0) {
-			    /* append failed */
-			    goto onError;
-			DPRINTF(" [%i:%i] (matched and remembered this slice)\n",
-				start,x);
-			}
-		    }
-		    else
-			DPRINTF(" [%i:%i] (matched but not saved)\n",start,x);
-		}
-		goto next_entry;
-	    }
-	    else
-		Py_ErrorWithArg(PyExc_TypeError,
-				"Tag Table entry %i: "
-				"expected a table (cmd=SubTable)",i);
+				"expected a table (cmd=[Sub]Table)",i);
 
 	    goto next_entry;
 	    
 	case MATCH_TABLEINLIST:
+	case MATCH_SUBTABLEINLIST:
 	    
 	    if (PyTuple_Check(match)) {
 		PyObject *subtags;
@@ -799,28 +744,27 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 				    "Tag Table entry %i: "
 				    "expected (list,index) as arg",i);
 		
-		if (taglist != Py_None) { 
-		    /* want a list, so create one */
+		if (taglist != Py_None && cmd != MATCH_SUBTABLE) {
+		    /* Create a new list for use as subtaglist */
 		    subtags = PyList_New(0);
-		    if (subtags == NULL) {
-			Py_DECREF(match);
+		    if (subtags == NULL) 
 			goto onError;
-		    }
 		}
 		else {
-		    /* better without list */
-		    subtags = Py_None;
+		    /* Use taglist as subtaglist */
+		    subtags = taglist;
 		    Py_INCREF(subtags);
 		}
 
-		DPRINTF("\nTableInList : using table at 0x%lx\n",(long)match);
+		DPRINTF("\n[Sub]TableInList : using table at 0x%lx\n",
+			(long)match);
 	    
 		start = x;
 
 		/* match other table */
 		newrc = fast_tag(pytext,text,len_text,match,start,subtags,&y);
 		if (newrc == 0) {
-		    Py_MY_DELETE(subtags);
+		    Py_DECREF(subtags);
 		    Py_DECREF(match);
 		    goto onError;
 		}
@@ -830,7 +774,9 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 		    DPRINTF(" (no success)\n");
 		    if (jne == 0) {
 			/* match failed */
-			rc = 1; 
+			rc = 1;
+			Py_DECREF(subtags);
+			Py_DECREF(match);
 			goto finished;
 		    }
 		    else 
@@ -846,7 +792,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 			if (match_append(flags,pytext,taglist,tagobj,
 					 start,x,subtags) < 0) {
 			    /* append failed */
-			    Py_MY_DELETE(subtags);
+			    Py_DECREF(subtags);
 			    Py_DECREF(match);
 			    goto onError;
 			}
@@ -857,7 +803,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 			DPRINTF(" [%i:%i] (matched but not saved)\n",start,x);
 		    }
 		}
-		Py_MY_DELETE(subtags);
+		Py_DECREF(subtags);
 		Py_DECREF(match);
 		goto next_entry;
 	    }
@@ -937,96 +883,36 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 	    goto next_entry;
 
 	case MATCH_CALL:
-	    
-	    if (PyCallable_Check(match)) {
-		PyObject *args;
-		register PyObject *w;
-
-		DPRINTF("\nCall :\n");
-
-		start = x;
-		
-		/* Build args = (pytext,start,len_text) */
-		args = PyTuple_New(3);
-		if (!args)
-		    goto onError;
-
-		Py_INCREF(pytext);
-		PyTuple_SET_ITEM(args,0,pytext);
-		    w = PyInt_FromLong(start);
-		    if (!w)
-			goto onError;
-		    PyTuple_SET_ITEM(args,1,w);
-		    w = PyInt_FromLong(len_text);
-		    if (!w)
-			goto onError;
-		    PyTuple_SET_ITEM(args,2,w);
-
-		w = PyEval_CallObject(match,args);
-		Py_DECREF(args);
-		if (w == NULL) 
-		    goto onError;
-
-		Py_AssertWithArg(PyInt_Check(w),
-				 PyExc_TypeError,
-				 "Tag Table entry %i: "
-				 "matching fct has to return an integer",i);
-		x = PyInt_AS_LONG(w);
-		Py_DECREF(w);
-
-		if (start == x) { 
-		    /* not matched */
-		    DPRINTF(" (no success)\n");
-		    if (jne == 0) {
-			/* match failed */
-			rc = 1; 
-			goto finished; 
-		    }
-		    else 
-			je = jne;
-		} 
-		else if (tagobj != Py_None) {
-		    /* matched */
-		    if (match_append(flags,pytext,taglist,tagobj,
-				     start,x,NULL) < 0)
-			goto onError;
-		    DPRINTF(" [%i:%i] (matched and remembered this slice)\n",
-			    start,x);
-		}
-		else {
-		    DPRINTF(" [%i:%i] (matched but not saved)\n",start,x);
-		}
-		goto next_entry;
-	    }
-	    else
-		Py_ErrorWithArg(PyExc_TypeError,
-				"Tag Table entry %i: "
-				"expected a callable object (cmd=Call)",i);
-
-	    goto next_entry;
-
 	case MATCH_CALLARG:
 	    {
 		PyObject *fct;
+		int argc;
 	    
-		if (!PyTuple_Check(match) || PyTuple_GET_SIZE(match) < 1)
-		    Py_ErrorWithArg(PyExc_TypeError,
+		if (!PyTuple_Check(match)) {
+		    argc = 0;
+		    fct = match;
+		}
+		else {
+		    argc = PyTuple_GET_SIZE(match) - 1;
+		    Py_AssertWithArg(argc >= 0,
+				     PyExc_TypeError,
 				    "Tag Table entry %i: "
 				    "expected a tuple (fct,arg0,arg1,...)"
 				    "(cmd=CallArg)",i);
-		fct = PyTuple_GET_ITEM(match,0);
+		    fct = PyTuple_GET_ITEM(match,0);
+		}
 	    
 		if (PyCallable_Check(fct)) {
 		    PyObject *args;
 		    register PyObject *w;
 		    register int i;
 
-		    DPRINTF("\nCallArg :\n");
+		    DPRINTF("\nCall[Arg] :\n");
 	    
 		    start = x;
 
-		    /* Build args = (pytext,start,len_text,arg0,arg1,...) */
-		    args = PyTuple_New(3 + PyTuple_GET_SIZE(match)-1);
+		    /* Build args = (pytext,start,len_text[,arg0,arg1,...]) */
+		    args = PyTuple_New(3 + argc);
 		    if (!args)
 			goto onError;
 		    Py_INCREF(pytext);
@@ -1039,10 +925,10 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 		    if (!w)
 			goto onError;
 		    PyTuple_SET_ITEM(args,2,w);
-		    for (i = 1; i < PyTuple_GET_SIZE(match); i++) {
-			w = PyTuple_GET_ITEM(match,i);
+		    for (i = 0; i < argc; i++) {
+			w = PyTuple_GET_ITEM(match,i + 1);
 			Py_INCREF(w);
-			PyTuple_SET_ITEM(args,2+i,w);
+			PyTuple_SET_ITEM(args,2 + i,w);
 		    }
 
 		    w = PyEval_CallObject(fct,args);
@@ -1086,7 +972,7 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 		    Py_ErrorWithArg(PyExc_TypeError,
 				    "Tag Table entry %i: "
 				    "expected a callable object "
-				    "(cmd=CallArg)",i);
+				    "(cmd=Call[Arg])",i);
 	    }
 	    goto next_entry;
 
@@ -1101,11 +987,24 @@ int fast_tag(PyObject *pytext,  	/* must be a Python string */
 
  finished:
 
-    DPRINTF("\nfinished: rc=%i; Tag Table entry %i; position %i\n",rc,i,x);
+    /* In case no specific return code was set, check if we have
+       matched successfully (pointer beyond the end of the table) or
+       failed to match (pointer negative) */
+    if (rc < 0) {
+	if (i >= len_table)
+	    rc = 2;
+	else if (i < 0)
+	    rc = 1;
+	else
+	    Py_ErrorWith2Args(PyExc_SystemError,
+			      "Internal Error: "
+			      "tagging engine finished with no proper result"
+			      "at position %i in table 0x%0x",
+			      i,(int)table);
+    }
 
-    /* Check if we have matched succesfully */
-    if (rc < 0 && i >= len_table)
-	rc = 2;
+    DPRINTF("\nTag Engine finished: %s; Tag Table entry %i; position %i\n",
+	    rc==1?"failed":"success",i,x);
 
     /* Record the current head position */
     *next = x;
@@ -1160,7 +1059,7 @@ int match_append(int flags,
 	/* Build w = (tagobj,l,r,subtags) */
 	w = PyTuple_New(4);
 	if (!w)
-	    return -1;
+	    goto onError;
 
 	Py_INCREF(tagobj);
 	PyTuple_SET_ITEM(w,0,tagobj);
@@ -1170,7 +1069,7 @@ int match_append(int flags,
 	PyTuple_SET_ITEM(w,3,subtags);
 
 	if (PyList_Append(taglist,w))
-	    return -1;
+	    goto onError;
 	Py_DECREF(w);
 	return 0;
     }
@@ -1184,6 +1083,21 @@ int match_append(int flags,
 	return PyList_Append(taglist,tagobj);
     }
 
+    if (flags & MATCH_APPENDMATCH) {
+	/* append the tagobj to the taglist */
+	register PyObject *v;
+	
+	if (taglist == Py_None) 
+	    return 0; /* nothing to be done */
+	v = PyString_FromStringAndSize(PyString_AS_STRING(pytext) + l, r-l);
+	if (!v)
+	    goto onError;
+	if (PyList_Append(taglist,v))
+	    goto onError;
+	Py_DECREF(v);
+	return 0;
+    }
+
     if (flags & MATCH_CALLTAG) { 
 	/* call tagobj */
 	register PyObject *args;
@@ -1191,7 +1105,7 @@ int match_append(int flags,
 	/* Build args = (taglist,pytext,l,r,subtags) */
 	args = PyTuple_New(5);
 	if (!args)
-	    return -1;
+	    goto onError;
 
 	Py_INCREF(taglist);
 	PyTuple_SET_ITEM(args,0,taglist);
@@ -1205,7 +1119,7 @@ int match_append(int flags,
 	w = PyEval_CallObject(tagobj,args);
 	Py_DECREF(args);
 	if (w == NULL)
-	    return -1;
+	    goto onError;
 	return 0;
     }
 
@@ -1218,7 +1132,7 @@ int match_append(int flags,
 	/* Build w = (None,l,r,subtags) */
 	w = PyTuple_New(4);
 	if (!w)
-	    return -1;
+	    goto onError;
 
 	Py_INCREF(Py_None);
 	PyTuple_SET_ITEM(w,0,Py_None);
@@ -1228,7 +1142,7 @@ int match_append(int flags,
 	PyTuple_SET_ITEM(w,3,subtags);
 
 	if (PyList_Append(tagobj,w))
-	    return -1;
+	    goto onError;
 	Py_DECREF(w);
 	return 0;
     }
